/**
  Raytracing demo application.
  
  Orion Sky Lawlor, olawlor@acm.org, 2010-01-22 (Public Domain)
*/
#include <iostream>
#include <cmath>
#include <GL/glew.h> /* 
OpenGL Extensions Wrangler:  for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include "ogl/glsl.h"
#include "ogl/glsl.cpp"
#include "ogl/minicam.h"

void display(void) 
{
	//std::cout<<"Called display function\n";
	glClearColor(0.4,0.5,0.7,0.0); // background color
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	
	glEnable(GL_DEPTH_TEST);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); // flush any ancient matrices
	gluPerspective(70.0, // FOV
		glutGet(GLUT_WINDOW_WIDTH)/float(glutGet(GLUT_WINDOW_HEIGHT)),
		0.1,100.0); // near and far clipping planes
	oglCameraLookAt(1.0); // look from camera
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity(); // flush any ancient matrices
	
	/* Draw labeled XYZ axes */
	oglCameraAxes();
	
	/* Draw a simple ground plane (to keep us oriented) */
	glEnable(GL_LINE_SMOOTH);
	glLineWidth(0.2);
	glColor3f(0,0,0);
	glBegin(GL_LINES);
	float size=10;
	for (float d=-size;d<=size;d++) {
		float z=-1.0;
		glVertex3f(d,-size,z);
		glVertex3f(d,+size,z);
		glVertex3f(-size,d,z);
		glVertex3f(+size,d,z);
	}
	glEnd();
	
	/* The proxy sphere should be gently rotating */
	const float rotate_rate=0.1; /* revolutions per second */
	float angle=glutGet(GLUT_ELAPSED_TIME)*0.001*rotate_rate*360;
	glRotatef(angle,0,1,0);
	
	/* Build a pixel shader */
	static GLhandleARB prog=MAKE_PROGRAM_OBJECT(
		// GLSL vertex shader!
		varying vec3 G; // proxy geometry location, world coordinates
		void main(void) {
			vec4 mv=gl_ModelViewMatrix*gl_Vertex; // apply modelview to get to world coords
			G=vec3(mv);
			gl_Position = gl_ProjectionMatrix*mv;
		}
	,
		// GLSL fragment shader
		uniform vec3 C; // camera's location, world coordinates
		varying vec3 G; // proxy geometry location, world coordinates
		void main(void) {
			vec3 D = normalize(G-C); // points from camera to geometry
			
			/*
			  ray equation:
			  	vec3 P = C + t * D;
			  equation of a paraboloid:
			  	k + P.z^2 = P.x^2 + P.y^2;
			  substituting:
			  	(C.z+t*D.z)^2 = (C.x + t*D.x)^2 + (C.y + t*D.y)^2
				Generally: (C + t D)^2 = C*C + 2*t*C*D + t*t*D*D
				k + C.z*C.z + 2*t*C.z*D.z + t*t*D.z*D.z
				  = (ditto, x) + (ditto,y)
				
				Or a quadratic with:
					a = D.z*D.z-D.x*D.x-D.y*D.y
					b = 2.0*(C.z*D.z - C.x*D.x - C.y*D.y)
					c = k+C.z*C.z - C.x*C.x - C.y*C.y
				and then solve the quadratic equation
			*/
			float k=-0.1;
			float a=D.z*D.z-D.x*D.x-D.y*D.y;
			float b = 2.0*(C.z*D.z - C.x*D.x - C.y*D.y);
			float c = k+C.z*C.z - C.x*C.x - C.y*C.y;
			float det=b*b-4.0*a*c;
			if (det<0) discard; /* miss! */
			float t=(-b+sqrt(det))/(2.0*a);
			if (t<0.0) discard; /* ray hits, but behind camera */
			vec3 P=C+t*D;
			if (length(P)>0.8) discard; /* outside bounding volume */
			
			/* Draw the hit location using "fract" wraparound plaid */
			gl_FragColor = vec4(fract(10.0*P),1.0);
		}
	);
	glUseProgramObjectARB(prog);
	glFastUniform3fv(prog,"C",1,camera); // set C to camera location (from ogl/minicam.h)
	glColor3f(0,0,1);
	glutSolidSphere(1.0,8,6); // proxy geometry
	glUseProgramObjectARB(0);
	
	glutPostRedisplay(); // continual animation
	glutSwapBuffers();
}

int main(int argc,char *argv[]) 
{
	glutInit(&argc,argv);
	glutInitDisplayMode(GLUT_RGBA + GLUT_DEPTH + GLUT_DOUBLE);
	glutInitWindowSize(800,600);
	glutCreateWindow("Test GL!");
	
	glewInit();
	
	glutDisplayFunc(display);
	oglCameraInit();
	camera=vec3(0,-3,0);
	
	glutMainLoop();
	return 0;
}
